@@ -339,6 +339,15 @@ void run(CuTest *tc, const char *format, ...) {
     free(command);
 }
 
+int should_run(const char *name, int argc, char **argv) {
+    if (argc == 0)
+        return 1;
+    for (int i=0; i < argc; i++)
+        if (STREQ(argv[i], name))
+            return 1;
+    return 0;
+}
+
 /*
  * Local variables:
  *  indent-tabs-mode: nil
